#!!! keep this file in project?

# Don't import numpy until after threads are set
import os

if False:
    thread_count = 12
    os.environ["MKL_NUM_THREADS"] = str(
        thread_count
    )  # Set this before numpy is imported
    os.environ["OPENBLAS_NUM_THREADS"] = str(thread_count)
    os.environ["OMP_NUM_THREADS"] = str(thread_count)
    os.environ["NUMEXPR_NUM_THREADS"] = str(thread_count)
    os.environ["VECLIB_MAXIMUM_THREADS"] = str(thread_count)
from pathlib import Path
import time
import logging
import pysnptools.util as pstutil
import platform
import multiprocessing

if False:
    cache_top = Path(r"c:\deldir")
    output_dir = Path(r"c:\users\carlk\OneDrive\Projects\Science\gpu_pref")
else:
    if platform.system() == "Windows":
        cache_top = Path(r"m:\deldir")
        output_dir = Path(r"d:\OneDrive\Projects\Science\gpu_pref")
    else:
        if "WSL_DISTRO_NAME" in os.environ:
            cache_top = Path(r"/mnt/m/deldir")
            output_dir = Path(r"/mnt/d/OneDrive\Projects\Science\gpu_pref_linux")
        else:
            cache_top = Path("/home/azureuser/deldir")
            output_dir = Path("/home/azureuser/gpu_pref_linux")


def one_experiment(
    test_snps,
    K0_goal,
    seed,
    pheno,
    covar,
    leave_out_one_chrom,
    use_gpu,
    proc_count,
    GB_goal,
    just_one_process=False,
    cpu_weight=1,
    gpu_weight=1,
    gpu_count=1,
    test_case="?",
    force=None,
    two_ks=False,
):
    import numpy as np
    from pysnptools.util.mapreduce1.runner import LocalMultiProc
    from unittest.mock import patch
    from fastlmm.association import single_snp

    if K0_goal is not None:
        K0 = test_snps[:, :: test_snps.sid_count // K0_goal]
    else:
        K0 = None

    if proc_count == 1:
        runner = None
        xp = "cupy" if use_gpu > 0 else "numpy"
    else:
        if use_gpu == 0:
            runner = LocalMultiProc(proc_count, just_one_process=just_one_process)
            xp = "numpy"
        else:
            assert gpu_count <= proc_count
            weights = [gpu_weight] * gpu_count + [cpu_weight] * (proc_count - gpu_count)

            def taskindex_to_environ(taskindex):
                if taskindex < gpu_count:
                    return {"ARRAY_MODULE": "cupy", "GPU_INDEX": str(taskindex)}
                else:
                    return {"ARRAY_MODULE": "numpy"}

            xp = "cupy"
            runner = LocalMultiProc(
                proc_count,
                weights=weights,
                taskindex_to_environ=taskindex_to_environ,
                just_one_process=just_one_process,
            )

    force_full_rank = False
    force_low_rank = False
    if force == "full_rank":
        force_full_rank = True
    elif force == "low_rank":
        force_low_rank = True
    elif force is None:
        pass
    else:
        assert False

    start_time = time.time()

    results_dataframe = single_snp(
        K0=K0,
        K1=test_snps if two_ks else None,
        test_snps=test_snps,
        pheno=pheno,
        covar=covar,
        leave_out_one_chrom=leave_out_one_chrom,
        count_A1=False,
        GB_goal=GB_goal,
        runner=runner,
        xp=xp,
        force_full_rank=force_full_rank,
        force_low_rank=force_low_rank,
    )
    delta_time = time.time() - start_time

    K0_count = test_snps.iid_count if K0 is None else K0.sid_count

    perf_result = {
        "computer_name": os.environ.get("COMPUTERNAME", "<unknown>"),
        "cpu_count": multiprocessing.cpu_count(),
        "gpu_count": "1?",
        "test_case": test_case,
        "linked": "MKL?/OpenBLAS?",
        "test_snps": str(test_snps),
        "iid_count": test_snps.iid_count,
        "test_sid_count": test_snps.sid_count,
        "low_rank": 0 if test_snps.iid_count == K0_count else 1,
        "K0_count": K0_count,
        "seed": seed,
        "chrom_count": len(np.unique(test_snps.pos[:, 0])),
        "covar_count": covar.col_count,
        "leave_out_one_chrom": 1 if leave_out_one_chrom else 0,
        "num_threads": os.environ.get("MKL_NUM_THREADS", "<none>"),
        "use_gpu": use_gpu,
        "cpu_weight": cpu_weight,
        "gpu_weight": gpu_weight,
        "proc_count": proc_count,
        "just_one_process": 1 if just_one_process else 0,
        "GB_goal": GB_goal,
        "time (s)": delta_time,
        "force": force if force is not None else "<none>",
        "two_ks": 1 if two_ks else 0,
    }
    return perf_result


def snpsA(seed, iid_count, sid_count, use_distributed):
    import numpy as np
    from pysnptools.snpreader import Bed
    from pysnptools.snpreader import DistributedBed
    from pysnptools.snpreader import SnpGen

    chrom_count = 10
    global top_cache
    if use_distributed:
        test_snp_path = (
            cache_top / f"snpsA_{seed}_{chrom_count}_{iid_count}_{sid_count}_db"
        )
    else:
        test_snp_path = (
            cache_top / f"snpsA_{seed}_{chrom_count}_{iid_count}_{sid_count}.bed"
        )
    count_A1 = False
    if not test_snp_path.exists():
        snpgen = SnpGen(
            seed=seed,
            iid_count=iid_count,
            sid_count=sid_count,
            chrom_count=chrom_count,
            block_size=1000,
        )
        if use_distributed:
            test_snps = DistributedBed.write(str(test_snp_path), snpgen)
        else:
            test_snps = Bed.write(
                str(test_snp_path), snpgen.read(dtype="float32"), count_A1=count_A1
            )
    else:
        if use_distributed:
            test_snps = DistributedBed(str(test_snp_path))
        else:
            test_snps = Bed(str(test_snp_path), count_A1=count_A1)
    from pysnptools.snpreader import SnpData

    np.random.seed(seed)
    pheno = SnpData(
        iid=test_snps.iid,
        sid=["pheno"],
        val=np.random.randn(test_snps.iid_count, 1) * 3 + 2,
    )
    covar = SnpData(
        iid=test_snps.iid,
        sid=["covar1", "covar2"],
        val=np.random.randn(test_snps.iid_count, 2) * 2 - 3,
    )

    return test_snps, pheno, covar


def pd_write(short_output_pattern, pref_list):
    import pandas as pd

    assert "{0}" in short_output_pattern, "Expect '{0}' in short_output_pattern"

    pref_df = pd.DataFrame(pref_list)
    global output_dir
    output_pattern = output_dir / short_output_pattern
    output_pattern.parent.mkdir(parents=True, exist_ok=True)
    i = 0
    while True:
        file = Path(str(output_pattern).format(i))
        if not file.exists():
            break
        i += 1
    pref_df.to_csv(file, sep="\t", index=False)
    print(pref_df)
    print(file)


def test_exp_1():
    short_output_pattern = "exp1/exp_1_{0}.tsv"

    # Set these as desired
    os.environ["MKL_NUM_THREADS"] = "1"  # Set this before numpy is imported
    seed = 1
    iid_count = 1 * 1000  # number of individuals
    sid_count = 5 * 1000  # number of SNPs
    K0_goal = 200
    leave_out_one_chrom = False

    # Tune these
    proc_count = 1
    GB_goal = 2

    test_snps, pheno, covar = snpsA(seed, iid_count, sid_count)

    pref_list = []
    for use_gpu in [False, True]:
        pref_list.append(
            one_experiment(
                test_snps,
                K0_goal,
                seed,
                pheno,
                covar,
                leave_out_one_chrom=leave_out_one_chrom,
                use_gpu=use_gpu,
                proc_count=proc_count,
                GB_goal=GB_goal,
            )
        )
    pd_write(short_output_pattern, pref_list)


def test_exp_2():
    short_output_pattern = "exp2/exp_2_{0}.tsv"

    # Set these as desired
    os.environ["MKL_NUM_THREADS"] = "20"  # Set this before numpy is imported
    seed = 1
    iid_count = 10 * 1000  # number of individuals
    sid_count = 50 * 1000  # number of SNPs
    K0_goal = 500
    leave_out_one_chrom = False

    # Tune these
    GB_goal = 2

    test_snps, pheno, covar = snpsA(seed, iid_count, sid_count)

    pref_list = []
    for proc_count, use_gpu in [(5, False), (10, False), (1, True), (2, True)]:
        logging.info(f"proc_count={proc_count},use_gpu={use_gpu}")
        pref_list.append(
            one_experiment(
                test_snps,
                K0_goal,
                seed,
                pheno,
                covar,
                leave_out_one_chrom=leave_out_one_chrom,
                use_gpu=use_gpu,
                proc_count=proc_count,
                GB_goal=GB_goal,
            )
        )
        pd_write(short_output_pattern + ".temp", pref_list)
    pd_write(short_output_pattern, pref_list)


def test_exp_3(K0_goal=500, GB_goal=2):
    short_output_pattern = f"exp3/exp_3_{K0_goal}_{'{0}'}.tsv"

    # Set these as desired
    os.environ["MKL_NUM_THREADS"] = "20"  # Set this before numpy is imported
    seed = 1
    iid_count = 10 * 1000  # number of individuals
    sid_count = 50 * 1000  # number of SNPs
    leave_out_one_chrom = True

    # Tune these

    test_snps, pheno, covar = snpsA(seed, iid_count, sid_count)

    pref_list = []
    for proc_count, use_gpu in [(1, True), (10, False)]:
        logging.info(f"proc_count={proc_count},use_gpu={use_gpu}")
        pref_list.append(
            one_experiment(
                test_snps,
                K0_goal,
                seed,
                pheno,
                covar,
                leave_out_one_chrom=leave_out_one_chrom,
                use_gpu=use_gpu,
                proc_count=proc_count,
                GB_goal=GB_goal,
            )
        )
        pd_write(short_output_pattern + ".temp", pref_list)
    pd_write(short_output_pattern, pref_list)


def testX_exp_4x(K0_goal=500):
    short_output_pattern = f"exp4x/exp_4x_{K0_goal}_{0}.tsv"

    # Set these as desired
    os.environ["MKL_NUM_THREADS"] = "1"  # Set this before numpy is imported
    seed = 1
    iid_count = 10 * 1000  # number of individuals
    sid_count = 50 * 1000  # number of SNPs

    leave_out_one_chrom = True

    # Tune these
    GB_goal = 2

    test_snps, pheno, covar = snpsA(seed, iid_count, sid_count)

    pref_list = []
    for proc_count, use_gpu in [(1, True), (1, False)]:
        logging.info(f"proc_count={proc_count},use_gpu={use_gpu}")
        pref_list.append(
            one_experiment(
                test_snps,
                K0_goal,
                seed,
                pheno,
                covar,
                leave_out_one_chrom=leave_out_one_chrom,
                use_gpu=use_gpu,
                proc_count=proc_count,
                GB_goal=GB_goal,
            )
        )
        pd_write(short_output_pattern + ".temp", pref_list)
    pd_write(short_output_pattern, pref_list)


def test_std():
    from pysnptools.standardizer.standardizer import Standardizer

    seed = 1
    iid_count = 2 * 1000
    sid_count = 50 * 1000  # number of SNPs
    test_snps, pheno, covar = snpsA(seed, iid_count, sid_count)
    xp = pstutil.array_module("cupy")
    print(xp.__name__)
    snps = xp.asarray(test_snps.read().val)
    stats = xp.empty(
        [snps.shape[1], 2],
        dtype=snps.dtype,
        order="F" if snps.flags["F_CONTIGUOUS"] else "C",
    )

    for i in range(15):
        Standardizer._standardize_unit_python(
            snps, apply_in_place=True, use_stats=False, stats=stats
        )
    print("done!")


def test_exp_4(
    GB_goal=2,
    iid_count=2 * 1000,
    sid_count=50 * 1000,
    K0_goal=None,
    proc_count_only_cpu=10,
    proc_count_with_gpu=5,
    cpu_weight=1,
    gpu_weight=1,
    gpu_count=1,
    num_threads=None,
    leave_out_one_chrom=True,
    just_one_process=False,
    force=None,
    two_ks=False,
    use_distributed=False,
):
    short_output_pattern = f"exp4/exp_4_{'{0}'}.tsv"

    # Set these as desired
    if num_threads is not None:
        os.environ["MKL_NUM_THREADS"] = str(
            num_threads
        )  # Set this before numpy is imported
        os.environ["OPENBLAS_NUM_THREADS"] = str(num_threads)
        os.environ["OMP_NUM_THREADS"] = str(num_threads)
        os.environ["NUMEXPR_NUM_THREADS"] = str(num_threads)
        os.environ["VECLIB_MAXIMUM_THREADS"] = str(num_threads)
    seed = 1

    # Tune these

    test_snps, pheno, covar = snpsA(
        seed, iid_count, sid_count, use_distributed=use_distributed
    )

    proc_gpu_list = []
    if proc_count_only_cpu > 0:
        proc_gpu_list += [(proc_count_only_cpu, 0)]
    if proc_count_with_gpu > 0:
        proc_gpu_list += [(proc_count_with_gpu, 0.5)]
    proc_gpu_list += [(1, 1)]
    pref_list = []
    for proc_count, use_gpu in proc_gpu_list:
        logging.info(f"proc_count={proc_count},use_gpu={use_gpu}")
        pref_list.append(
            one_experiment(
                test_snps,
                K0_goal=K0_goal,
                seed=seed,
                pheno=pheno,
                covar=covar,
                leave_out_one_chrom=leave_out_one_chrom,
                use_gpu=use_gpu,
                proc_count=proc_count,
                GB_goal=GB_goal,
                just_one_process=just_one_process,
                cpu_weight=cpu_weight,
                gpu_weight=gpu_weight,
                gpu_count=gpu_count,
                force=force,
                two_ks=two_ks,
            )
        )
        pd_write(short_output_pattern + ".temp", pref_list)
    pd_write(short_output_pattern, pref_list)


def test_case_def(test_case):
    if test_case == "a":
        iid_count = 2000
        sid_count = 50 * 1000
        K0_goal = None
    elif test_case == "b":
        iid_count = 4000
        sid_count = 50 * 1000
        K0_goal = None
    elif test_case == "c":
        iid_count = 5000
        sid_count = 50 * 1000
        K0_goal = 2000
    elif test_case == "d":
        iid_count = 1000
        sid_count = 100 * 1000
        K0_goal = 1000
    elif test_case == "e":
        iid_count = 2000
        sid_count = 1000 * 1000
        K0_goal = None
    elif test_case == "f":
        iid_count = 200
        sid_count = 200
        K0_goal = None
    else:
        assert False, test_case

    return test_case, iid_count, sid_count, K0_goal


def test_svd(size, which_list, threads=None):
    # may want to play with "F" vs "C" order
    if threads is not None:
        os.environ["MKL_NUM_THREADS"] = str(threads)
    import numpy as np

    short_output_pattern = f"svd/svd_{'{0}'}.tsv"
    which_list = which_list or [
        "np.linalg.svd",
        "cp.linalg.svd",
    ]
    m_row = size
    n_col = m_row + 2
    min_row_col = size
    perf_list = []
    for which in which_list:
        logging.info("generating {0}x{1}".format(m_row, n_col))
        np.random.seed(0)
        a = np.random.randn(m_row, n_col).astype(np.float)  # what's the "float about?"
        if which == "np.linalg.svd":
            xp = np
            logging.info(f"Doing large np.linalg.svd ({m_row}x{n_col})")
            start_time = time.time()
            U, s, V = np.linalg.svd(a, full_matrices=False, compute_uv=True)
            delta_time = time.time() - start_time
            logging.info(
                f"done with np.linalg.svd {m_row}x{n_col} in time {delta_time}"
            )
            print(U.shape, s.shape, V.shape, end=" ")
            S = np.zeros((min_row_col, min_row_col))
            S = np.diag(s)
            assert np.allclose(a, np.dot(U, np.dot(S, V)))
        elif which == "cp.linalg.svd":
            import cupy as cp

            xp = cp
            logging.info(f"Doing large cp.linalg.svd ({m_row}x{n_col})")
            a = cp.asarray(a)
            start_time = time.time()
            U, s, V = cp.linalg.svd(a, full_matrices=False, compute_uv=True)
            delta_time = time.time() - start_time
            logging.info(
                f"done with cp.linalg.svd {m_row}x{n_col} in time {delta_time}"
            )
            print(U.shape, s.shape, V.shape, end=" ")
            S = cp.zeros((min_row_col, min_row_col))
            S = cp.diag(s)
            assert cp.allclose(a, cp.dot(U, cp.dot(S, V)))
        else:
            assert False
        perf_result = {
            "computer_name": os.environ.get("COMPUTERNAME", "<unknown>"),
            "cpu_count": multiprocessing.cpu_count(),
            "gpu_count": "1?",
            "test_case": "test_svd",
            "linked": "MKL?/OpenBLAS?",
            "size": size,
            "which": which,
            "num_threads": os.environ.get("MKL_NUM_THREADS", "<none>"),
            "xp": xp.__name__,
            "time (s)": delta_time,
        }
        perf_list.append(perf_result)
        pd_write(short_output_pattern + ".temp", perf_list)
    pd_write(short_output_pattern, perf_list)


def benchmark1():
    # based on https://towardsdatascience.com/heres-how-to-use-cupy-to-make-numpy-700x-faster-4b920dda1f56
    import numpy as np
    import cupy as cp
    import time

    ### Numpy and CPU
    s = time.time()
    x_cpu = np.ones((500, 1000, 1000))
    e = time.time()
    print(e - s)

    ### CuPy and GPU
    s = time.time()
    x_gpu = cp.ones((500, 1000, 1000))
    # cp.cuda.Stream.null.synchronize()
    e = time.time()
    print(e - s)

    ### Numpy and CPU
    s = time.time()
    x_cpu *= 5
    e = time.time()
    print(e - s)
    ### CuPy and GPU
    s = time.time()
    x_gpu *= 5
    cp.cuda.Stream.null.synchronize()
    e = time.time()
    print(e - s)

    print("done")


def benchmark2():
    # Based on http://learningsys.org/nips17/assets/papers/paper_16.pdf
    import time, numpy

    # import cupy

    def test1(xp, size, repeat):
        a = xp.arange(size).reshape(1000, -1)
        # cupy.cuda.runtime.deviceSynchronize()
        t1 = time.time()
        for index in range(repeat):
            a = a.T * 2
        # cupy.cuda.runtime.deviceSynchronize()
        t2 = time.time()
        diff = t2 - t1
        print(f"{repeat}x{size:,} {xp.__name__} {diff}")
        return max(t2 - t1, 1e-16)

    def test(xp, size, repeat):
        a = xp.arange(size).reshape(1000, -1)
        b = xp.arange(size).reshape(-1, 1000)
        # cupy.cuda.runtime.deviceSynchronize()
        t1 = time.time()
        for index in range(repeat):
            a = xp.dot(b, a)
        # cupy.cuda.runtime.deviceSynchronize()
        t2 = time.time()
        diff = t2 - t1
        print(f"{repeat}x{size:,} {xp.__name__} {diff}")
        return max(t2 - t1, 1e-16)

    repeat = 50
    for exp in range(6, 9):
        size = 10**exp
        t_dict = {}
        for xp in [numpy]:
            t_dict[xp.__name__ + "1st"] = test(xp, size, 1)  # Avoid first call overhead
            t_dict[xp.__name__ + "2nd"] = test(xp, size, 1)
        # print(f"{repeat}x{size:,} 1st:{t_dict['numpy1st']/t_dict['cupy1st']} 2nd:{t_dict['numpy2nd']/t_dict['cupy2nd']}")
    # (’numpy’, 0.5105748176574707)
    # (’cupy’, 0.08418107032775879)


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.INFO)

    if True:
        test_case, iid_count, sid_count, K0_goal = test_case_def("d")

        for two_ks in [False]:
            for force in ["full_rank", "low_rank"]:
                test_exp_4(
                    GB_goal=4,
                    iid_count=iid_count,
                    K0_goal=K0_goal,
                    proc_count_only_cpu=0,
                    proc_count_with_gpu=6,
                    cpu_weight=4,
                    gpu_weight=3,
                    gpu_count=2,
                    num_threads=None,
                    leave_out_one_chrom=True,
                    just_one_process=False,
                    force=force,
                    two_ks=two_ks,
                    use_distributed=True,
                )

    # benchmark2()

    # test_svd(3_000, which_list=None)



import numpy as NP
from numpy import dot
import scipy.stats as ST
from scipy.linalg import cholesky, solve_triangular
from fastlmm.external.util.math import (
    check_definite_positiveness,
    check_symmetry,
    mvnormpdf,
    ddot,
    trace2,
)
from fastlmm.external.util.math import stl, stu, dotd
from fastlmm.inference.glmm import GLMM_N1K3, GLMM_N3K1
from fastlmm.inference.likelihood import LogitLikelihood, ProbitLikelihood
from fastlmm import Pr
from . import likelihood as LH


"""
    Important! Always run test.py in the current folder for unit testing after
    changes have been made.
"""


class EPGLMM(object):
    def __init__(self, link):
        self._debugUACalls = []
        self._link = link
        if link == "logistic":
            self._likelihood = LogitLikelihood()
        elif link == "erf":
            self._likelihood = ProbitLikelihood()
        else:
            assert False, "Unknown link function."

    def _updateApproximation(self):
        """
        Calculates the Laplace approximation for the posterior.
        It can be defined by two variables: f mode and W at f mode.
        """
        if self._updateApproximationCount == 0:
            return

        if self._is_kernel_zero():
            self._updateApproximationCount = 0
            return

        self._updateApproximationBegin()

        self._mean = self._calculateMean()
        m = self._mean

        ttau = NP.zeros(self._N)
        tnu = NP.zeros(self._N)
        sig2_ = NP.zeros(self._N)
        mu_ = NP.zeros(self._N)

        prevsig2 = self._dKn()

        prevmu = m.copy()

        converged = False
        outeriter = 1
        iterMax = 1000

        while outeriter <= iterMax and not converged:
            tau_ = 1.0 / prevsig2 - ttau
            nu_ = prevmu / prevsig2 - tnu

            tt = tau_**2 + tau_
            stt = NP.sqrt(tt)
            c = (self._y * nu_) / stt

            # dan: using _cdf and _pdf instead of cdf and pdf I avoid
            # a lot of overhead due to error checking and other things
            nc_hz = NP.exp(ST.norm._logpdf(c) - ST.norm._logcdf(c))

            hmu = nu_ / tau_ + nc_hz * self._y / stt

            hsig2 = 1.0 / tau_ - (nc_hz / tt) * (c + nc_hz)

            tempttau = 1.0 / hsig2 - tau_
            ok = NP.bitwise_and(~NP.isnan(tempttau), abs(tempttau) > 1e-8)
            ttau[ok] = tempttau[ok]

            tnu[ok] = hmu[ok] / hsig2[ok] - nu_[ok]

            (sig2, mu) = self._calculateSig2Mu(ttau, tnu, m)

            lastMuDiff = abs(mu - prevmu).max()
            lastSig2Diff = abs(sig2 - prevsig2).max()
            if lastMuDiff < 1e-6 and lastSig2Diff < 1e-6:
                converged = True

            prevmu = mu
            prevsig2 = sig2

            outeriter += 1

        if outeriter > iterMax:
            Pr.prin("EP did not converge. Printing debug information...")
            Pr.prin("beta " + str(self.beta))
            Pr.prin(
                "sig02 "
                + str(self.sig02)
                + " sig12 "
                + str(self.sig12)
                + " sign2 "
                + str(self.sign2)
            )
            Pr.prin("abs(mu-prevmu).max() " + str(lastMuDiff))
            Pr.prin("abs(sig2-prevsig2).max() " + str(lastSig2Diff))

        self._updateApproximationEnd(ttau, tnu, tau_, nu_)
        self._updateApproximationCount = 0


class EPGLMM_N1K3(GLMM_N1K3, EPGLMM):
    def __init__(self, link, penalty=None, penalizeBias=False, debug=False):
        GLMM_N1K3.__init__(
            self, penalty=penalty, penalizeBias=penalizeBias, debug=debug
        )
        EPGLMM.__init__(self, link)
        self._link = link

    def _updateApproximationBegin(self):
        self._G01 = self._calculateG01()

    def _updateApproximationEnd(self, ttau, tnu, tau_, nu_):
        self._ttau = ttau
        self._tnu = tnu
        self._tau_ = tau_
        self._nu_ = nu_
        self._V = self._calculateV(ttau, self._sign2)
        self._Lk = self._calculateLk(self._G01, self._V)
        self._H = stl(self._Lk, ddot(self._G01.T, self._V, left=False))

    def _regular_marginal_loglikelihood(self):
        self._updateConstants()
        self._updateApproximation()

        if self._is_kernel_zero():
            return self._likelihood.log(self._mean, self._y)

        m = self._mean
        sign2 = self._sign2
        tnu = self._tnu
        ttau = self._ttau
        tau_ = self._tau_
        nu_ = self._nu_
        G01 = self._G01
        V = self._V
        Lk = self._Lk
        H = self._H

        A = ttau * sign2 + 1.0
        Ktnu = self._rdotK(tnu)
        HKtnu = dot(H, Ktnu)
        Hm = dot(H, m)
        ts_1 = 1.0 / (tau_ + ttau)
        mV = m * V

        r = (
            0.5 * NP.sum(NP.log(1.0 + ttau / tau_))
            - 0.5 * NP.sum(NP.log(A))
            - NP.sum(NP.log(NP.diag(Lk)))
            + NP.sum(ST.norm._logcdf((self._y * nu_) / NP.sqrt(tau_**2 + tau_)))
            + 0.5
            * (
                dot(tnu, Ktnu)
                - dot(Ktnu.T, V * Ktnu)
                + dot(HKtnu.T, HKtnu)
                - dot(tnu, ts_1 * tnu)
            )
            + dot(m, tnu)
            - dot(mV, Ktnu)
            + dot(Hm.T, HKtnu)
            - 0.5 * (dot(mV, m) - dot(Hm.T, Hm))
            + 0.5 * dot(nu_ * ts_1, (ttau / tau_) * nu_ - 2.0 * tnu)
        )

        return r

    def _rmll_gradient(self, optSig02=True, optSig12=True, optSign2=True, optBeta=True):
        self._updateConstants()
        self._updateApproximation()

        m = self._mean
        H = self._H
        V = self._V
        ttau = self._ttau
        tnu = self._tnu
        G0 = self._G0
        G1 = self._G1

        Smtnu = ttau * m - tnu
        KSmtnu = self._rdotK(Smtnu)

        b = Smtnu - V * KSmtnu + dot(H.T, dot(H, KSmtnu))

        ret = []
        if optSig02:
            r = 0.5 * (
                dot(b, dot(G0, dot(G0.T, b)))
                - trace2(ddot(V, G0, left=True), G0.T)
                + trace2(H.T, dot(dot(H, G0), G0.T))
            )

            ret.append(r)

        if optSig12:
            r = 0.5 * (
                dot(b, dot(G1, dot(G1.T, b)))
                - trace2(ddot(V, G1, left=True), G1.T)
                + trace2(H.T, dot(dot(H, G1), G1.T))
            )

            ret.append(r)

        if optSign2:
            r = 0.5 * (dot(b, b) - NP.sum(V) + trace2(H.T, H))

            ret.append(r)

        if optBeta:
            ret += list(-dot(b, self._X))

        return NP.array(ret)

    def _updateApproximation(self):
        EPGLMM._updateApproximation(self)

    def _calculateV(self, ttau, sign2):
        V = ttau / (ttau * sign2 + 1.0)
        return V

    def _calculateSig2Mu(self, ttau, tnu, mean):
        sign2 = self._sign2

        V = self._calculateV(ttau, sign2)
        G01 = self._G01

        Lk = self._calculateLk(G01, V)

        G01tV = ddot(G01.T, V, left=False)
        H = stl(Lk, G01tV)

        HK = self._ldotK(H)

        sig2 = (
            self._dKn()
            - sign2**2 * V
            - dotd(G01, dot(dot(G01tV, G01), G01.T))
            - 2.0 * sign2 * dotd(G01, G01tV)
            + dotd(HK.T, HK)
        )

        assert NP.all(NP.isfinite(sig2)), "sig2 should be finite."

        u = self._mean + self._rdotK(tnu)

        mu = u - self._rdotK(V * u) + self._rdotK(H.T.dot(H.dot(u)))
        assert NP.all(NP.isfinite(mu)), "mu should be finite."

        return (sig2, mu)

    def _predict(self, meanstar, kstar, kstarstar, prob):
        self._updateConstants()
        self._updateApproximation()

        m = self._mean
        tnu = self._tnu
        Lk = self._Lk
        H = self._H
        V = self._V

        Ktnu = self._rdotK(tnu)
        mKtnu = m + Ktnu
        Vkstar = ddot(V, kstar, left=True)
        Hkstar = dot(H, kstar)

        mustar = (
            meanstar
            + dot(kstar.T, tnu)
            - dot(Vkstar.T, mKtnu)
            + dot(Hkstar.T, dot(H, mKtnu))
        )

        if prob is False:
            if nom > 0.0:
                return +1.0
            return -1.0

        sig2star = (
            kstarstar
            - dotd(kstar.T, Vkstar)
            + dotd(dot(Hkstar.T, H), kstar)
            + self._sign2
        )

        return LH.probit_sigmoid(mustar / NP.sqrt(1.0 + sig2star))


class EPGLMM_N3K1(GLMM_N3K1, EPGLMM):
    def __init__(self, link, penalty=None, penalizeBias=False, debug=False):
        GLMM_N3K1.__init__(
            self, penalty=penalty, penalizeBias=penalizeBias, debug=debug
        )
        EPGLMM.__init__(self, link)
        self._link = link

    def _updateApproximationBegin(self):
        self._K = NP.eye(self._N) * self._sign2
        if self._isK0Set:
            self._K += self._sig02 * (
                self._K0 if self._K0 is not None else dot(self._G0, self._G0.T)
            )
        if self._isK1Set:
            self._K += self._sig12 * (
                self._K1 if self._K1 is not None else dot(self._G1, self._G1.T)
            )

    def _regular_marginal_loglikelihood(self):
        self._updateConstants()
        self._updateApproximation()

        if self._is_kernel_zero():
            return self._likelihood.log(self._mean, self._y)

        m = self._mean
        K = self._K
        Ln = self._Ln
        ttau = self._ttau
        tnu = self._tnu
        tau_ = self._tau_
        nu_ = self._nu_
        LnSsq = self._LnSsq
        LnSsqK = self._LnSsqK

        LnSsqKtnu = dot(LnSsqK, tnu)
        ts_1 = 1.0 / (tau_ + ttau)
        LnSsqm = dot(LnSsq, m)

        r = (
            0.5 * NP.sum(NP.log(1.0 + ttau / tau_))
            - NP.sum(NP.log(NP.diag(Ln)))
            + NP.sum(ST.norm._logcdf((self._y * nu_) / NP.sqrt(tau_**2 + tau_)))
            + 0.5
            * (dot(tnu, dot(K, tnu)) - dot(LnSsqKtnu, LnSsqKtnu) - dot(tnu * ts_1, tnu))
            + dot(m, tnu)
            - dot(LnSsqm.T, LnSsqKtnu)
            - 0.5 * dot(LnSsqm.T, LnSsqm)
            + 0.5 * dot(nu_ * ts_1, (ttau / tau_) * nu_ - 2.0 * tnu)
        )

        return r

    def _rmll_gradient(self, optSig02=True, optSig12=True, optSign2=True, optBeta=True):
        self._updateConstants()
        self._updateApproximation()

        K0 = self._K0
        K1 = self._K1
        m = self._mean
        ttau = self._ttau
        tnu = self._tnu
        Ssq = self._Ssq
        LnSsq = self._LnSsq
        LnSsqK = self._LnSsqK

        Smtnu = ttau * m - tnu

        b = Smtnu - dot(LnSsq.T, dot(LnSsqK, Smtnu))

        ret = []

        if optSig02:
            r = 0.5 * (dot(b, dot(K0, b)) - trace2(LnSsq.T, dot(LnSsq, K0)))
            ret.append(r)

        if optSig12:
            r = 0.5 * (dot(b, dot(K1, b)) - trace2(LnSsq.T, dot(LnSsq, K1)))
            ret.append(r)

        if optSign2:
            r = 0.5 * (dot(b, b) - trace2(LnSsq.T, LnSsq))
            ret.append(r)

        if optBeta:
            r = -dot(b, self._X)
            ret += list(r)

        ret = NP.array(ret)
        assert NP.all(
            NP.isfinite(ret)
        ), "Not finite regular marginal loglikelihood gradient."

        return ret

    def _updateApproximation(self):
        EPGLMM._updateApproximation(self)

    def _updateApproximationEnd(self, ttau, tnu, tau_, nu_):
        self._ttau = ttau
        self._tnu = tnu
        self._tau_ = tau_
        self._nu_ = nu_
        self._Ssq = NP.sqrt(ttau)
        self._Ln = self._calculateLn(self._K, self._Ssq)
        self._LnSsq = stl(self._Ln, NP.diag(self._Ssq))
        self._LnSsqK = dot(self._LnSsq, self._K)

    def _calculateSig2Mu(self, ttau, tnu, mean):
        Ssq = NP.sqrt(ttau)

        Ln = self._calculateLn(self._K, Ssq)

        LnSsq = stl(Ln, NP.diag(Ssq))
        V = dot(LnSsq, self._K)
        sig2 = self._dKn() - dotd(V.T, V)
        mu = (
            mean
            + dot(self._K, tnu)
            - dot(V.T, dot(LnSsq, mean))
            - dot(V.T, dot(V, tnu))
        )

        return (sig2, mu)

    def _predict(self, meanstar, kstar, kstarstar, prob):
        self._updateConstants()
        self._updateApproximation()

        m = self._mean
        tnu = self._tnu
        K = self._K
        Ln = self._Ln
        Ssq = self._Ssq
        LnSsq = self._LnSsq
        LnSsqkstar = dot(LnSsq, kstar)
        LnSsqK = self._LnSsqK

        mustar = (
            meanstar
            + dot(kstar.T, tnu)
            - dot(LnSsqkstar.T, dot(LnSsq, m) + dot(LnSsqK, tnu))
        )

        if prob is False:
            if nom > 0.0:
                return +1.0
            return -1.0

        sig2star = kstarstar - dotd(dot(LnSsqkstar.T, LnSsq), kstar) + self._sign2

        return LH.probit_sigmoid(mustar / NP.sqrt(1.0 + sig2star))



import numpy as np
import subprocess, sys, os.path
from itertools import *
from fastlmm.pyplink.snpset import *
import logging


class SnpAndSetNameCollection(object):  # implements ISnpSetList
    """
    Specifies a list of snp sets via a file that has columns 'snp' and 'group'.
    See the Bed class's 'read' method of examples of its use.
    See __init__.py for specification of interface it implements.
    """

    def __init__(self, filename):
        self.filename = filename
        logging.info("Reading {0}".format(filename))
        import pandas as pd

        snp_and_setname_sequence = pd.read_csv(
            filename, delimiter="\s", index_col=False, engine="python"
        )

        from collections import defaultdict

        setname_to_snp_list = defaultdict(list)
        for snp, gene in snp_and_setname_sequence.itertuples(index=False):
            setname_to_snp_list[gene].append(snp)
        self.bigToSmall = sorted(
            setname_to_snp_list.items(),
            key=lambda gene_snp_list: -len(gene_snp_list[1]),
        )

    def addbed(self, bed):
        return SnpAndSetNameCollectionPlusBed(self, bed)

    def copyinputs(self, copier):
        copier.input(self.filename)

    # would be nicer if these used generic pretty printer
    def __repr__(self):
        return "SnpAndSetNameCollection(filename={0})".format(self.filename)

    def __iter__(self):
        for gene, snp_list in self.bigToSmall:
            yield gene, snp_list


class GenomeSplitCollection(object):  # implements ISnpSetList
    """
    Specifies a list of snp sets by splitting the genome into windows of fixed size.
    """

    def __init__(self, bed, windowsize, idist):  # implements ISnpSetListPlusBed
        """
        Input:
        bed         : bed object
        windowsize  : size of the window
        idist       : index in pos array that the exclusion is based on.
                  (1=genetic distance, 2=basepair distance)
        """
        self.bed = bed
        self.snp_list = []
        chrom = bed.pos[:, 0]
        dist = bed.pos[:, idist]

        # divide into blocks
        chromUnique = np.unique(chrom)
        blockId = 0
        for chromId in chromUnique:
            chromIdx = chromId == chrom
            distStart = dist[chromIdx].min()
            distMax = dist[chromIdx].max()

            while distStart < distMax:
                distStop = distStart + windowsize
                idxBlock = np.logical_and(
                    chromIdx, np.logical_and(distStart <= dist, dist < distStop)
                )
                self.snp_list.append(("block%d" % blockId, bed.rs[idxBlock]))
                blockId += 1
                distStart = distStop

    def __len__(self):
        return len(self.snp_list)

    def __iter__(self):
        for block, snp_list in self.snp_list:
            if len(set(snp_list)) != len(snp_list):
                raise Exception("Some snps are listed more than once.")
            yield SnpAndSetNamePlusBed(block, snp_list, self.bed)


class SnpAndSetNameCollectionPlusBed(object):  # implements ISnpSetListPlusBed
    """
    The SnpAndSetNameCollection with the addition of BED information.
    """

    def __init__(self, spec, bed):
        self.spec = spec
        self.bed = bed

    def __len__(self):
        return len(self.spec.bigToSmall)

    def __iter__(self):
        for gene, snp_list in self.spec.bigToSmall:
            if len(set(snp_list)) != len(snp_list):
                raise Exception(
                    "Some snps in gene {0} are listed more than once".format(gene)
                )
            yield SnpAndSetNamePlusBed(gene, snp_list, self.bed)



import doctest
import unittest
import numpy as np
import logging
import os
import pandas as pd
import tempfile
import shutil
import sys
from numpy.random import RandomState

import pysnptools.util as pstutil
from pysnptools.snpreader import Bed, DistributedBed
from pysnptools.snpreader import Pheno, SnpData
from pysnptools.util.filecache import LocalCache
from pysnptools.util.filecache import PeerToPeer
from pysnptools.snpreader import SnpGen

from fastlmm.association import single_snp
from fastlmm.association import single_snp_scale
from pysnptools.util.mapreduce1.runner import LocalMultiProc


class TestSingleSnpScale(unittest.TestCase):
    @classmethod
    def test_snpgen(self):
        seed = 0
        snpgen = SnpGen(seed=seed, iid_count=1000, sid_count=5000, block_size=100)
        snpdata = snpgen[:, [0, 1, 200, 2200, 10]].read()
        np.testing.assert_allclose(
            np.nanmean(snpdata.val, axis=0),
            np.array(
                [
                    0.0013089005235602095,
                    0.0012953367875647669,
                    0.014084507042253521,
                    0.0012422360248447205,
                    0.0012674271229404308,
                ]
            ),
            rtol=1e-5,
        )

        snpdata2 = snpgen[:, [0, 1, 200, 2200, 10]].read()
        np.testing.assert_equal(snpdata.val, snpdata2.val)
        snpdata3 = snpgen[::10, [0, 1, 200, 2200, 10]].read()
        np.testing.assert_equal(snpdata3.val, snpdata2.val[::10, :])

    def test_snpgen_cache(self):
        cache_file = tempfile.gettempdir() + "/test_snpgen_cache.snpgen.npz"
        if os.path.exists(cache_file):
            os.remove(cache_file)
        snpgen = SnpGen(
            seed=0,
            iid_count=1000,
            sid_count=5000,
            cache_file=cache_file,
            block_size=100,
        )
        assert os.path.exists(cache_file)
        snpgen2 = SnpGen(
            seed=0,
            iid_count=1000,
            sid_count=5000,
            cache_file=cache_file,
            block_size=100,
        )
        os.remove(cache_file)
        snpdata = snpgen2[:, [0, 1, 200, 2200, 10]].read()
        np.testing.assert_allclose(
            np.nanmean(snpdata.val, axis=0),
            np.array(
                [
                    0.0013089005235602095,
                    0.0012953367875647669,
                    0.014084507042253521,
                    0.0012422360248447205,
                    0.0012674271229404308,
                ]
            ),
            rtol=1e-5,
        )

    @classmethod
    def setUpClass(self):
        from pysnptools.util import create_directory_if_necessary
        import fastlmm as fastlmm

        create_directory_if_necessary(self.tempout_dir, isfile=False)
        self.pythonpath = os.path.abspath(
            os.path.join(os.path.dirname(os.path.realpath(fastlmm.__file__)), "..")
        )
        self.bed = Bed(
            os.path.join(self.pythonpath, "tests/datasets/synth/all.bed"), count_A1=True
        )[:, ::10]
        self.phen_fn = os.path.join(
            self.pythonpath, "tests/datasets/synth/pheno_10_causals.txt"
        )
        self.cov_fn = os.path.join(self.pythonpath, "tests/datasets/synth/cov.txt")

    tempout_dir = "tempout/single_snp_scale"

    def test_old(self):
        logging.info("test_old")

        output_file = self.file_name("old")
        results_df = single_snp(
            self.bed,
            self.phen_fn,
            covar=self.cov_fn,
            count_A1=True,
            output_file_name=output_file,
        )
        self.compare_files(results_df, "old")

    @staticmethod
    def _cache_dict(top, clear_cache):
        storage = top.join("intermediate")
        if clear_cache:
            storage.rmtree()
        cache_dict = {chrom: storage for chrom in range(23)}
        return cache_dict

    def test_low(self):
        logging.info("test_low")

        output_file = self.file_name("low")

        storage = LocalCache("local_cache/low")
        for clear_cache in (True, False):
            if clear_cache:
                storage.rmtree()
            results_df = single_snp_scale(
                test_snps=self.bed,
                pheno=self.phen_fn,
                covar=self.cov_fn,
                cache=storage,
                output_file_name=output_file,
            )
            self.compare_files(results_df, "old")

    def test_multipheno(self):
        logging.info("test_multipheno")

        random_state = RandomState(29921)
        pheno_reference = Pheno(self.phen_fn).read()
        for pheno_count in [2, 5, 1]:
            val = random_state.normal(
                loc=pheno_count,
                scale=pheno_count,
                size=(pheno_reference.iid_count, pheno_count),
            )
            pheno_col = ["pheno{0}".format(i) for i in range(pheno_count)]
            pheno_multi = SnpData(iid=pheno_reference.iid, sid=pheno_col, val=val)

            reference = pd.concat(
                [
                    single_snp(
                        test_snps=self.bed,
                        pheno=pheno_multi[:, pheno_index],
                        covar=self.cov_fn,
                    )
                    for pheno_index in range(pheno_count)
                ]
            )
            frame = single_snp_scale(
                test_snps=self.bed, pheno=pheno_multi, covar=self.cov_fn
            )

            assert len(frame) == len(
                reference
            ), "# of pairs differs from file '{0}'".format(reffile)
            for sid in sorted(
                set(reference.SNP)
            ):  # This ignores which pheno produces which pvalue
                pvalue_frame = np.array(sorted(frame[frame["SNP"] == sid].PValue))
                pvalue_reference = np.array(
                    sorted(reference[reference["SNP"] == sid].PValue)
                )
                assert (
                    abs(pvalue_frame - pvalue_reference) < 1e-5
                ).all, "pair {0} differs too much from reference".format(sid)

    def test_local_distribute(self):  # cmk
        logging.info("test_local_distribute")
        force_python_only = False

        output_file = self.file_name("local_distribute")

        storage = LocalCache("local_cache/local_distribute")
        test_storage = storage.join("test_snps")
        test_storage.rmtree("")
        test_snps = DistributedBed.write(
            test_storage, self.bed, piece_per_chrom_count=2
        )

        results_df = single_snp_scale(
            test_snps=test_snps,
            pheno=self.phen_fn,
            covar=self.cov_fn,
            G0=self.bed,
            cache=self._cache_dict(storage, clear_cache=True),
            output_file_name=output_file,
            force_python_only=force_python_only,
        )

        self.compare_files(results_df, "old")

        results_df = single_snp_scale(
            test_snps=self.bed,
            pheno=self.phen_fn,
            covar=self.cov_fn,
            G0=self.bed,
            cache=self._cache_dict(storage, clear_cache=False),
            output_file_name=output_file,
        )
        self.compare_files(results_df, "old")

    def test_fast_mapreduce1_runner(self):
        logging.info("test_fast_mapreduce1_runner")

        output_file = self.file_name("mapreduce1_runner")
        runner = LocalMultiProc(taskcount=4, just_one_process=True)
        single_snp_scale(
            test_snps=self.bed[:, ::100],
            pheno=self.phen_fn,
            covar=self.cov_fn,
            output_file_name=output_file,
            memory_factor=100,
            runner=runner,
        )

    def test_mapreduce1_runner(self):
        logging.info("test_mapreduce1_runner")

        output_file = self.file_name("mapreduce1_runner")
        runner = LocalMultiProc(taskcount=4, just_one_process=True)
        results_df = single_snp_scale(
            test_snps=self.bed,
            pheno=self.phen_fn,
            covar=self.cov_fn,
            output_file_name=output_file,
            runner=runner,
        )
        self.compare_files(results_df, "old")

    def test_old_one(self):
        logging.info("test_old_one")

        output_file = self.file_name("old_one")

        test_snps3 = self.bed[:, self.bed.pos[:, 0] == 3]  # Test only on chromosome 3
        results_df = single_snp(
            test_snps=test_snps3,
            K0=self.bed,
            pheno=self.phen_fn,
            covar=self.cov_fn,
            count_A1=True,
            output_file_name=output_file,
        )
        self.compare_files(results_df, "old_one")

    def test_one_fast(self):
        logging.info("test_one_fast")

        output_file = self.file_name("one_fast")

        storage = LocalCache("local_cache")
        test_storage = storage.join("one_fast")
        test_storage.rmtree()
        test_snps3 = self.bed[:, self.bed.pos[:, 0] == 3]  # Test only on chromosome 3
        test_snps3_dist = DistributedBed.write(
            test_storage, test_snps3, piece_per_chrom_count=2
        )

        results_df = single_snp_scale(
            test_snps=test_snps3_dist,
            pheno=self.phen_fn,
            covar=self.cov_fn,
            G0=self.bed,
            output_file_name=output_file,
        )
        self.compare_files(results_df, "old_one")

    def test_one_chrom(self):
        logging.info("test_one_chrom")

        output_file = self.file_name("one_chrom")

        storage = LocalCache("local_cache/one_chrom")
        test_storage = storage.join("test_snps")
        test_storage.rmtree("")
        test_snps3 = self.bed[:, self.bed.pos[:, 0] == 3]  # Test only on chromosome 3
        test_snps3_dist = DistributedBed.write(
            test_storage, test_snps3, piece_per_chrom_count=2
        )

        for test_snps, ref, clear_cache, name in (
            (test_snps3, "old_one", True, "Run with just chrom3"),
            (test_snps3_dist, "old_one", True, "Run with distributed test SNPs"),
            (test_snps3, "old_one", False, "Run with just chrom3 (use cache)"),
            (
                test_snps3_dist,
                "old_one",
                False,
                "Run with distributed test SNPs (use cache)",
            ),
        ):
            logging.info("=========== " + name + " ===========")
            results_df = single_snp_scale(
                test_snps=test_snps,
                pheno=self.phen_fn,
                covar=self.cov_fn,
                K0=self.bed,
                cache=self._cache_dict(storage, clear_cache=clear_cache),
                output_file_name=output_file,
            )
            self.compare_files(results_df, ref)

    def too_slow_test_peertopeer(self):
        logging.info("test_peertopeer")

        output_file = self.file_name("peertopeer")

        def id_and_path_function():
            from pysnptools.util.filecache import ip_address_pid

            ip_pid = ip_address_pid()
            # Need to put the 'cache_top' here explicitly.
            return ip_pid, "peertopeer/{0}".format(ip_pid)

        storage = PeerToPeer(
            common_directory="peertopeer/common",
            id_and_path_function=id_and_path_function,
        )
        test_snps_cache = storage.join("test_snps")
        test_snps_cache.rmtree()
        test_snps = DistributedBed.write(
            test_snps_cache, self.bed, piece_per_chrom_count=2
        )

        runner = LocalMultiProc(taskcount=5)  # Run on 5 additional Python processes

        for clear_cache in (True, False):
            if clear_cache:
                storage.join("cache").rmtree()
            results_df = single_snp_scale(
                test_snps=test_snps,
                pheno=self.phen_fn,
                covar=self.cov_fn,
                cache=storage.join("cache"),
                output_file_name=output_file,
                runner=runner,
            )
            self.compare_files(results_df, "old")

    def file_name(self, testcase_name):
        temp_fn = os.path.join(self.tempout_dir, testcase_name + ".txt")
        if os.path.exists(temp_fn):
            os.remove(temp_fn)
        return temp_fn

    def compare_files(self, frame, ref_base):
        reffile = self.reference_file("single_snp_scale/" + ref_base + ".txt")

        reference = pd.read_csv(reffile, delimiter="\s", comment=None, engine="python")
        assert len(frame) == len(
            reference
        ), "# of pairs differs from file '{0}'".format(reffile)
        for _, row in reference.iterrows():
            sid = row.SNP
            pvalue = frame[frame["SNP"] == sid].iloc[0].PValue
            assert (
                abs(row.PValue - pvalue) < 1e-5
            ), "pair {0} differs too much from file '{1}'".format(sid, reffile)

    @staticmethod
    def reference_file(outfile):
        #!!similar code elsewhere
        import platform

        os_string = platform.platform()

        file_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "..", "..", "..", "tests"
        )
        windows_fn = file_path + "/expected-Windows/" + outfile
        assert os.path.exists(windows_fn), "Can't find file '{0}'".format(windows_fn)
        debian_fn = file_path + "/expected-debian/" + outfile
        if not os.path.exists(
            debian_fn
        ):  # If reference file is not in debian folder, look in windows folder
            debian_fn = windows_fn

        if "debian" in os_string or "Linux" in os_string:
            if "Linux" in os_string:
                logging.warning(
                    "comparing to Debian output even though found: %s" % os_string
                )
            return debian_fn
        else:
            if "Windows" not in os_string:
                logging.warning(
                    "comparing to Windows output even though found: %s" % os_string
                )
            return windows_fn

    def test_doctest(self):  # Can't get doc test to work so marked out.
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/..")
        doctest.ELLIPSIS_MARKER = "-etc-"
        result = doctest.testmod(
            sys.modules["fastlmm.association.single_snp_scale"],
            optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE,
        )
        doctest.ELLIPSIS_MARKER = "..."
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():
    """
    set up composite test suite
    """
    test_suite = unittest.TestSuite([])
    test_suite.addTests(
        unittest.TestLoader().loadTestsFromTestCase(TestSingleSnpScale)
    )  # Tests Ludicrous Speed GWAS
    return test_suite


if __name__ == "__main__":
    # logging.basicConfig(level=logging.INFO)
    logging.getLogger().setLevel(logging.INFO)
    suites = getTestSuite()

    if True:
        r = unittest.TextTestRunner(failfast=False)
        ret = r.run(suites)
        assert ret.wasSuccessful()
    else:  # runner test run
        logging.basicConfig(level=logging.INFO)

        from pysnptools.util.mapreduce1.distributabletest import DistributableTest

        runner = LocalMultiProc(taskcount=22, mkl_num_threads=5, just_one_process=False)
        distributable_test = DistributableTest(suites, "temp_test")
        print(runner.run(distributable_test))

    logging.info("done")



import numpy as np
import logging


class Unit(object):  # IStandardizer
    """The specification for unit standardization"""

    def __init__(self):
        import warnings

        #!!warnings.warn("This Unit is deprecated. Pysnptools includes newer versions of Unit", DeprecationWarning)
        pass

    def standardize(
        self, snps, blocksize=None, force_python_only=False, num_threads=None
    ):
        l = self.lambdaFactory(
            snps,
            blocksize=blocksize,
            force_python_only=force_python_only,
            num_threads=num_threads,
        )
        import fastlmm.util.standardizer as stdizer

        return stdizer.standardize_with_lambda(snps, l, blocksize)

    @staticmethod
    def _standardizer(snps, force_python_only, num_threads):
        from pysnptools.standardizer import Standardizer

        Standardizer._standardize_unit_and_beta(
            snps,
            is_beta=False,
            a=float("NaN"),
            b=float("NaN"),
            apply_in_place=True,
            use_stats=False,
            stats=None,
            force_python_only=force_python_only,
            num_threads=num_threads,
        )
        return snps

    def lambdaFactory(
        self, snps, blocksize=None, force_python_only=False, num_threads=None
    ):
        return lambda s, force_python_only=force_python_only: self._standardizer(
            snps, force_python_only, num_threads
        )

    def __str__(self):
        return "{0}()".format(self.__class__.__name__)



import fastlmm.association.lrt as lr
import numpy as np
import fastlmm.util.stats.chi2mixture as c2
import fastlmm.association.score as score
import scipy.linalg as LA
import scipy.stats as ST
from . import tests_util as tu


class Sc(object):
    """description of class"""

    __slots__ = ["score"]

    def check_nperm(self, nperm):
        return 0  # ignore request for permuations

    def __init__(self, score):
        self.score = score

    def __str__(self):
        return "sc_{0}".format(self.score)

    def construct(
        self,
        Y,
        X=None,
        forcefullrank=False,
        SNPs0=None,
        i_exclude=None,
        nullModel=None,
        altModel=None,
        scoring=None,
        greater_is_better=None,
    ):
        """
        The same code gets executed for both linear and logistic, because the logistic is an approximation.
        """
        assert (
            nullModel["effect"] == "mixed" and altModel["effect"] == "mixed"
        ), "You have not used mixed effects for the two kernel case."

        assert nullModel["link"] == "linear"
        assert altModel["link"] == "linear"

        if self.score == "mom":
            assert nullModel["link"] == "linear" and altModel["link"] == "linear", (
                "You are allowed to use " "only the linear link for sc mom test."
            )
        G0, K0 = tu.set_snps0(
            SNPs0=SNPs0,
            sample_size=Y.shape[0],
            i_exclude=i_exclude,
            forcefullrank=forcefullrank,
        )
        return score.scoretest2K(Y=Y[:, np.newaxis], X=X, K=K0, G0=G0)

    def construct_no_backgound_kernel(
        self, Y, X, forcefullrank, nullModel, altModel, scoring, greater_is_better
    ):
        if nullModel["link"] == "logistic":
            assert (
                nullModel["effect"] == "fixed"
                and altModel["link"] == "logistic"
                and altModel["effect"] == "mixed"
            ), (
                "By choosing the logistic link for the "
                "null model, you should also select "
                "fixed effect for null model, "
                "mixed effect and logistic link for alt "
                "model."
            )
            return score.scoretest_logit(Y=Y[:, np.newaxis], X=X)

        elif nullModel["link"] == "linear":
            assert (
                nullModel["effect"] == "fixed"
                and altModel["link"] == "linear"
                and altModel["effect"] == "mixed"
            ), (
                "By choosing the linear link for the "
                "null model, you should also select "
                "fixed effect for null model, "
                "mixed effect and linear link for alt "
                "model."
            )
            return score.scoretest(Y=Y[:, np.newaxis], X=X)
        else:
            raise Exception("Invalid link for score null model: " + link)

    def pv(self, squaredform, expectationsqform, varsqform, GPG):
        if self.score == "davies":
            return Sc.pv_davies(squaredform, expectationsqform, varsqform, GPG)
        elif self.score == "mom":
            return Sc.pv_mom(squaredform, expectationsqform, varsqform, GPG)
        else:
            raise Exception("Don't know sc score " + self.score)

    @staticmethod
    def pv_davies_eig(squaredform, eigvals):
        import fastlmmclib.quadform as qf

        # result = qf.qf(squaredform, eigvals,acc=1e-04,lim=10000)    #settings to match R-based results
        result = qf.qf(
            squaredform, eigvals, acc=1e-07
        )  # decided on 1e-7 after experimentation between -4 and -12. Thresh on exp in QFC.C seems to have no effect
        return result[0]

    @staticmethod
    def pv_davies(squaredform, expectationsqform, varsqform, GPG):
        eigvals = LA.eigh(GPG, eigvals_only=True)
        pv = Sc.pv_davies_eig(squaredform, eigvals)
        return pv

    @staticmethod
    def pv_mom(squaredform, expectationsqform, varsqform, GPG):
        """
        Do moment of matching on the scaled chi square approximation to the null, as in Li and Cui
        """
        dofchi2 = 2.0 * varsqform * expectationsqform * expectationsqform
        scalechi2 = 0.5 / (expectationsqform * varsqform)
        pv = ST.chi2.sf(squaredform / scalechi2, dofchi2)
        return pv

    @property
    def npvals(self):
        return 1  # return the 1 type of p-Value

    def w2(self, G0, result):
        return 0.0

    def lrt_method(self, result):
        return 0.0

    def pv_adj_from_result(self, result):
        return result.pv

    def pv_adj_and_ind(
        self,
        nperm,
        pv_adj,
        nullfit,
        lrt,
        lrtperm,
        alteqnull,
        alteqnullperm,
        qmax,
        nullfitfile,
        nlocalperm,
    ):
        ind = pv_adj.flatten().argsort()
        return pv_adj, ind

    def write(self, fp, ind, result_dict, pv_adj, detailed_table):
        fp.write(
            "\t".join(
                (
                    "SetId",
                    "P-value",
                    "#SNPs_in_Set",
                    "#ExcludedSNPs",
                    "test stat",
                    "chrm",
                    "pos. range",
                )
            )
            + "\n"
        )
        for i in range(len(ind)):
            ii = ind[i]
            result = result_dict[ii]
            fp.write(
                "\t".join(
                    (
                        result.setname,
                        str(pv_adj[ii]),
                        str(result.setsize),
                        str(result.nexclude),
                        str(result.stat),
                        result.ichrm,
                        result.iposrange + "\n",
                    )
                )
            )

    def pv_etc(
        self, filenull, G0_to_use, G1, y, x, null_model, varcomp_test, forcefullrank
    ):
        if forcefullrank:
            raise Exception("full rank score not implemented here")
        if self.filenull is not None:
            pv, garbage = score.twokerneltest(
                G0=G0_to_use,
                G1=G1,
                y=y,
                nulldistrib=self,
                covar=x,
                appendbias=False,
                test2K=null_model,
            )
            return pv, np.nan, np.nan
        else:
            pv = score.onekerneltest(
                G0=G1, y=y, nulldistrib=self, covar=x, appendbias=False
            )
            return pv, np.nan, np.nan



import numpy as np
import subprocess, sys, os.path
from itertools import *
from fastlmm.pyplink.snpset import *
import math


class Consecutive(object):  # implements ISnpSetList
    """
    The sets should be every consecutive set of SNPs within a 2cM window of each user
    (distance in cM is in the 3rd column of the bim file).  As for the name of the set,
    please make it <position-of-first-snp>@<position-of-middle-snp>@<position-of-last-snp>.
    For 'middle' please break a tie to the first SNP.
    """

    def __init__(self, bimFileName, cMWindow):
        self.BimFileName = bimFileName
        self.CMWindow = cMWindow

    def addbed(self, bed):
        return ConsecutivePlusBed(self, bed)

    def copyinputs(self, copier):
        copier.input(self.BimFileName)

    # would be nicer if these used generic pretty printer
    def __repr__(self):
        return "Consecutive(bimFileName={0},bimFileName={1})".format(
            self.BimFileName, self.CMWindow
        )


class ConsecutivePlusBed(object):  # implements ISnpSetListPlusBed
    def __init__(self, spec, bed):
        self.spec = spec
        self.bed = bed
        import pandas as pd

        bimfields = pd.read_csv(
            self.spec.BimFileName,
            delimiter="\s",
            usecols=(0, 1, 2, 3),
            header=None,
            index_col=False,
            engine="python",
        )
        self.chrom = bimfields[0]
        self.rs = bimfields[1]
        self.cm = bimfields[2]

    def __iter__(self):
        startIndex = -1
        endIndex = 0  # one too far
        while True:
            startIndex += 1
            if startIndex >= len(self.rs):
                return
            while (
                endIndex < len(self.rs)
                and self.chrom[startIndex] == self.chrom[endIndex]
                and self.cm[endIndex] - self.cm[startIndex] <= self.spec.CMWindow
            ):
                endIndex += 1
            lastIndex = endIndex - 1
            midIndex = math.floor((startIndex + lastIndex) / 2.0)
            name = "{0}@{1}@{2}".format(startIndex, midIndex, lastIndex)

            snpList = self.rs[list(range(startIndex, endIndex))]
            yield SnpAndSetNamePlusBed(name, snpList, self.bed)

    def __len__(self):
        return len(self.rs)



import pandas as pd
import sys
import numpy as np
import logging
import unittest
import doctest
import os.path
from unittest.mock import patch

from pysnptools.snpreader import Bed, Pheno,SnpData
from fastlmm.association import single_snp_all_plus_select,single_snp
from fastlmm.feature_selection.test import TestFeatureSelection
import multiprocessing
from pysnptools.util.mapreduce1.runner import Local, LocalMultiProc, LocalInParts, LocalMultiThread

class TestSingleSnpAllPlusSelect(unittest.TestCase): 

    @classmethod
    def setUpClass(self):
        from pysnptools.util import create_directory_if_necessary
        create_directory_if_necessary(self.tempout_dir, isfile=False)
        self.pythonpath = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)),"..","..",".."))
        self.bedbase = os.path.join(self.pythonpath, 'tests/datasets/all_chr.maf0.001.N300')
        self.phen_fn = os.path.join(self.pythonpath, 'tests/datasets/phenSynthFrom22.23.N300.randcidorder.txt')
        self.cov_fn = os.path.join(self.pythonpath,  'tests/datasets/all_chr.maf0.001.covariates.N300.txt')

    tempout_dir = "tempout/single_snp_all_plus_select"

    def file_name(self,testcase_name):
        temp_fn = os.path.join(self.tempout_dir,testcase_name+".txt")
        if os.path.exists(temp_fn):
            os.remove(temp_fn)
        return temp_fn

    def too_slow_test_notebook(self):
        do_plot = False
        runner = LocalMultiProc(multiprocessing.cpu_count(),mkl_num_threads=2)
        output_file_name = self.file_name("notebook")


        logging.info("TestSingleSnpAllPlusSelect test_notebook")
        # define file names
        snp_reader = Bed(self.pythonpath + "/tests/datasets/synth/all.bed",count_A1=False)
        pheno_fn = self.pythonpath + "/tests/datasets/synth/pheno_10_causals.txt"
        cov_fn = self.pythonpath + "/tests/datasets/synth/cov.txt"

        # find the chr5 SNPs
        test_snps = snp_reader[:,snp_reader.pos[:,0] == 5]

        #select the 2nd kernel and run GWAS
        results = single_snp_all_plus_select(test_snps=test_snps,G=snp_reader,pheno=pheno_fn,GB_goal=2,do_plot=do_plot,output_file_name=output_file_name,runner=runner, count_A1=False)


        self.compare_files(results,"notebook")

    def test_one(self):
        logging.info("TestSingleSnpAllPlusSelect test_one")
        snps = self.bedbase
        pheno = self.phen_fn
        covar = self.cov_fn
        
        output_file_name = self.file_name("one")
        results = single_snp_all_plus_select(test_snps=snps, pheno=pheno,
                                  covar=covar,
                                  k_list = np.logspace(start=0, stop=1, num=2, base=2),
                                  n_folds=2,
                                  do_plot=False,
                                  output_file_name = output_file_name,
                                  GB_goal=2,
                                  #runner = LocalMultiProc(taskcount=20,mkl_num_threads=5,just_one_process=True)
                                  count_A1=False
                                  )

        self.compare_files(results,"one")

    def too_slow_test_three(self):
        logging.info("TestSingleSnpAllPlusSelect test_three")

        bed_fn = self.pythonpath + "/tests/datasets/synth/all.bed"
        bed_fn = Bed(bed_fn,count_A1=False)
        pheno_fn = self.pythonpath + "/tests/datasets/synth/pheno_10_causals.txt"
        cov_fn = self.pythonpath + "/tests/datasets/synth/cov.txt"
        runner = LocalMultiProc(multiprocessing.cpu_count(),mkl_num_threads=2)

        output_file_name = self.file_name("three")
        results = single_snp_all_plus_select(test_snps=bed_fn, pheno=pheno_fn,
                                  covar=cov_fn,
                                  k_list = [int(k) for k in np.logspace(0, 7, base=2, num=7)],
                                  n_folds=7,
                                  seed = 42,
                                  do_plot=False,
                                  GB_goal=2,
                                  output_file_name=output_file_name,
                                  runner = runner,count_A1=False
                                  )
        logging.info(results)
        self.compare_files(results,"three")

    def too_slowtest_two(self): #!!! rather a big test case
        from pysnptools.util.mapreduce1.runner import Local, LocalMultiProc
        logging.info("TestSingleSnpAllPlusSelect test_two")
        do_plot = False

        bed_fn = self.pythonpath + "/tests/datasets/synth/all.bed"
        pheno_fn = self.pythonpath + "/tests/datasets/synth/pheno_10_causals.txt"
        cov_fn = self.pythonpath + "/tests/datasets/synth/cov.txt"

        # partition snps on chr5 vs rest
        test_chr = 5
        snp_reader = Bed(bed_fn,count_A1=False)
        test_snps = snp_reader[:,snp_reader.pos[:,0] == test_chr]
        runner = LocalMultiProc(multiprocessing.cpu_count(),mkl_num_threads=2)

        output_file_name = self.file_name("two")
        for GB_goal in [None,2]:
            results = single_snp_all_plus_select(test_snps=test_snps, G=bed_fn, pheno=pheno_fn,
                                      covar=cov_fn,
                                      k_list = [int(k) for k in np.logspace(0, 7, base=2, num=7)],
                                      n_folds=7,
                                      seed = 42,
                                      do_plot=do_plot,
                                      GB_goal=GB_goal,
                                      output_file_name=output_file_name,
                                      runner = runner,
                                      count_A1=False
                                      )
            logging.info(results.head())
            self.compare_files(results,"two")

    def test_old(self):
        with patch.dict('os.environ', {'ARRAY_MODULE': 'numpy'}) as _:

            do_plot = False
            from fastlmm.feature_selection.feature_selection_two_kernel import FeatureSelectionInSample
            from pysnptools.util import intersect_apply

            logging.info("TestSingleSnpAllPlusSelect test_old")

            bed_fn = self.pythonpath + "/tests/datasets/synth/all.bed"
            pheno_fn = self.pythonpath + "/tests/datasets/synth/pheno_10_causals.txt"
            cov_fn = self.pythonpath + "/tests/datasets/synth/cov.txt"

            #load data
            ###################################################################
            snp_reader = Bed(bed_fn,count_A1=False)
            pheno = Pheno(pheno_fn)
            cov = Pheno(cov_fn)

            # intersect sample ids
            snp_reader, pheno, cov = intersect_apply([snp_reader, pheno, cov])

            # read in snps

            # partition snps on chr5 vs rest
            test_chr = 5
            G0 = snp_reader[:,snp_reader.pos[:,0] != test_chr].read(order='C').standardize()
            test_snps = snp_reader[:,snp_reader.pos[:,0] == test_chr].read(order='C').standardize()


            y = pheno.read().val[:,0]
            y -= y.mean()
            y /= y.std()

            # load covariates
            X_cov = cov.read().val
            X_cov.flags.writeable = False

            # invoke feature selection to learn which SNPs to use to build G1
            logging.info("running feature selection conditioned on background kernel")
            # partition data into the first 50 SNPs on chr1 and all but chr1

        select = FeatureSelectionInSample(max_log_k=7, n_folds=7, order_by_lmm=True, measure="ll", random_state=42)
        best_k, feat_idx, best_mix, best_delta = select.run_select(G0.val, G0.val, y, cov=X_cov)    

        # plot out of sample error
        if do_plot: select.plot_results(measure="ll")
        # select.plot_results(measure="mse")

        # print results
        logging.info("best_k:{0}".format(best_k))
        logging.info("best_mix:{0}".format(best_mix))
        logging.info("best_delta:{0}".format(best_delta))


        ###############################
        # use selected SNPs to build G1
        logging.info(feat_idx)
        G1 = G0[:,feat_idx]

        output_file_name = self.file_name("old")
        results_df = single_snp(test_snps, pheno, G0=G0, G1=G1, mixing=best_mix, h2=None,leave_out_one_chrom=False,output_file_name=output_file_name,count_A1=False)

        logging.info("results:")
        logging.info("#"*40)
        logging.info(results_df.head())
        self.compare_files(results_df,"old")


    def compare_files(self,frame,ref_base):
        reffile = TestFeatureSelection.reference_file("single_snp_all_plus_select/"+ref_base+".txt")

        #sid_list,pvalue_list = frame['SNP'].values,frame['Pvalue'].values

        #sid_to_pvalue = {}
        #for index, sid in enumerate(sid_list):
        #    sid_to_pvalue[sid] = pvalue_list[index]

        reference=pd.read_csv(reffile,delimiter='\s',comment=None,engine='python')
        if 'Pvalue' in reference.columns: reference['PValue']=reference.Pvalue #add a new column with different capitalization if it is there


        assert len(frame) == len(reference), "# of pairs differs from file '{0}'".format(reffile)
        for _, row in reference.iterrows():
            sid = row.SNP
            pvalue = frame[frame['SNP'] == sid].iloc[0].PValue
            assert abs(row.PValue - pvalue) < 1e-5, "pair {0} differs too much from file '{1}'".format(sid,reffile)

    def too_slow_test_doctest(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__))+"/..")
        result = doctest.testmod(sys.modules['fastlmm.association.single_snp_all_plus_select'])
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__



def getTestSuite():
    
    suite1 = unittest.TestLoader().loadTestsFromTestCase(TestSingleSnpAllPlusSelect)
    return unittest.TestSuite([suite1])



if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    # this import is needed for the runner
    from fastlmm.association.tests.test_single_snp_all_plus_select import TestSingleSnpAllPlusSelect
    suites = unittest.TestSuite([getTestSuite()])

    if True: #Standard test run
        r = unittest.TextTestRunner(failfast=True)
        ret = r.run(suites)
        assert ret.wasSuccessful()
    else: #Cluster test run



        from pysnptools.util.mapreduce1.runner import Local, LocalMultiProc
        logging.basicConfig(level=logging.INFO)

        from pysnptools.util.mapreduce1.distributabletest import DistributableTest


        runner = Local()
        #runner = LocalMultiProc(taskcount=20,mkl_num_threads=5)
        #runner = LocalInParts(1,2,mkl_num_threads=1) # For debugging the cluster runs
        distributable_test = DistributableTest(suites,"temp_test")
        print(runner.run(distributable_test))


    logging.info("done with testing")



import os
import logging
import numpy as np
import datetime
import multiprocessing
from pysnptools.util.mapreduce1 import map_reduce
from pysnptools.kernelreader import KernelData, KernelNpz
import time
from pysnptools.util import format_delta
from pysnptools.snpreader import SnpMemMap
from bed_reader import file_dot_piece, file_b_less_aatbx, get_num_threads


def mmultfile_ata(memmap_lambda,writer,sid,work_count,name,runner,force_python_only=False):
    sid_count = len(sid)
    piece_count = work_count * 2
    log_frequency = 1 if logging.getLogger().level <= logging.INFO else 0

    def debatch_closure(piece_index):
        return sid_count * piece_index // piece_count

    def mapper_closure(work_index):
        memmap = memmap_lambda()
        piece_index0 = work_index
        piece_index1 = piece_count-work_index-1
        gtg_piece0 = mmultfile_ata_piece(memmap.filename,memmap.offset,piece_index0,piece_count,log_frequency=log_frequency,force_python_only=force_python_only)
        gtg_piece1 = mmultfile_ata_piece(memmap.filename,memmap.offset,piece_index1,piece_count,log_frequency=log_frequency,force_python_only=force_python_only)
        return [[piece_index0, gtg_piece0],[piece_index1, gtg_piece1]]

    def reducer_closure(result_result_sequence):
        logging.info("starting ata reducer")
        iid = [[value, value] for value in sid]
        gtg_data = KernelData(iid=iid, val=np.zeros((sid_count,sid_count)))
        for result_result in result_result_sequence:
            for piece_index, gtg_piece in result_result:
                logging.info("combining ata reducer {0}".format(piece_index))
                start = debatch_closure(piece_index)
                stop = debatch_closure(piece_index+1)
                gtg_data.val[start:,start:stop] = gtg_piece
                gtg_data.val[start:stop,start+gtg_piece.shape[1]:] = gtg_piece[gtg_piece.shape[1]:,:].T
        result = writer(gtg_data)
        return result

    gtg_npz_lambda = map_reduce(range(work_count),
               mapper=mapper_closure,
               reducer=reducer_closure,
               runner=runner,
               name=name,
               input_files=[],
               output_files=[]
               )

def mmultfile_ata_piece(a_filename, offset, work_index=0, work_count=1,log_frequency=-1, force_python_only=False):
    t0_gtg = time.time()
    if log_frequency > 0:
        logging.info("ata_piece: Working on piece {0} of {1}.".format(work_index, work_count))

    a = SnpMemMap(a_filename)

    def debatch_closure(work_index2):
        return a.sid_count * work_index2 // work_count
    start = debatch_closure(work_index)
    stop = debatch_closure(work_index+1)

    ata_piece = np.zeros((a.sid_count-start,stop-start),order='C')

    do_both = False
    if force_python_only or do_both:
        with open(a.filename,"rb") as fp:
            fp.seek(a.offset+start*a.iid_count*8)
            slice = np.fromfile(fp, dtype=np.float64, count=a.iid_count*(stop-start)).reshape(a.iid_count,stop-start,order="F")
            for i in range(work_index,work_count):
                starti = debatch_closure(i)
                stopi = debatch_closure(i+1)
                if i > work_index:
                    slicei = np.fromfile(fp, dtype=np.float64, count=a.iid_count*(stopi-starti)).reshape(a.iid_count,stopi-starti,order="F")
                else:
                    slicei = slice
                if log_frequency > 0 and i%log_frequency == 0:
                    logging.info("{0}/{1}".format(i,work_count))
                ata_piece[starti-start:stopi-start,:] = np.dot(slicei.T,slice)
        if do_both:
            ata_piece_python = ata_piece
            ata_piece = np.zeros((a.sid_count-start,stop-start),order='C')
    if not force_python_only or do_both:
        file_dot_piece(str(a_filename),a.offset,a.iid_count,
            start,
            ata_piece,
            num_threads=get_num_threads(None),
            log_frequency=log_frequency)

    if do_both:
        if not np.abs(ata_piece_python-ata_piece).max() < 1e-12:
           raise AssertionError("Expect Python and Rust to get the same file_dot answer")

    if log_frequency > 0:
        logging.info("ata_piece {0} of {1}: clocktime {2}".format(work_index, work_count,format_delta(time.time()-t0_gtg)))
    return ata_piece

def mmultfile_b_less_aatb(a_snp_mem_map, b, log_frequency=0, force_python_only=False):

    # Without memory efficiency
    #   a=a_snp_mem_map.val
    #   aTb = np.dot(a.T,b)
    #   aaTb = b-np.dot(a,aTb)
    #   return aTb, aaTb

    do_both = False

    if force_python_only or do_both:
        aTb = np.zeros((a_snp_mem_map.sid_count,b.shape[1])) #b can be destroyed. Is everything is in best order, i.e. F vs C
        aaTb = b.copy()
        b_mem = np.array(b,order="F") #!!! if we want this in "F" how about just creating that way instead of copying it
        with open(a_snp_mem_map.filename,"rb") as U_fp:
            U_fp.seek(a_snp_mem_map.offset)
            for i in range(a_snp_mem_map.sid_count):
                a_mem = np.fromfile(U_fp, dtype=np.float64, count=a_snp_mem_map.iid_count)
                if log_frequency > 0 and i%log_frequency == 0:
                    logging.info("{0}/{1}".format(i,a_snp_mem_map.sid_count))
                aTb[i,:] = np.dot(a_mem,b_mem)
                aaTb -= np.dot(a_mem.reshape(-1,1),aTb[i:i+1,:])
    if not force_python_only or do_both:
        if do_both:
            aTb_python = aTb
            aaTb_python = aaTb

        b1 = np.array(b,order="F")
        aTb = np.zeros((a_snp_mem_map.sid_count,b.shape[1]))
        aaTb = np.array(b1,order="F")

        file_b_less_aatbx(str(a_snp_mem_map.filename),
                            a_snp_mem_map.offset,
                            a_snp_mem_map.iid_count, #row count
                        b1,   #B copy 1 in "F" order
                        aaTb, #B copy 2 in "F" order
                        aTb, # result
                        num_threads = get_num_threads(None),
                        log_frequency=log_frequency,
            )
        
    if do_both:
        if not np.abs(aTb_python-aTb).max() < 1e-12 or not np.abs(aaTb_python-aaTb).max() < 1e-12:
           raise AssertionError("Expect Python and Rust to get the same mmultfile_b_less_aatb answer")
        
    return aTb, aaTb

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    test_postsvd = True
    test_ata = False
    assert test_postsvd+test_ata <= 1, "Can test at most one thing"


    do_fast = False
    do_medium = False
    do_mid16 = True
    assert do_fast+do_medium <=1, "Can test as most one speed (aka size)"


    if test_postsvd:
        if do_fast:
            filename = r"D:\deldir\test\_Storage\very_small_0.8\G0_data.memmap"
        elif do_medium:
            filename = r"D:\deldir\test\_Storage\mid_0.8\G0_data.memmap"
        elif do_mid16:
            filename = r"D:\deldir\test\_Storage\mid_1.6\G0_data.memmap"
        else:
            filename = r"D:\deldir\scratch\escience\carlk\cachebio\genetics\onemil\fc\bigsyn0\U1.memmap"
        

        do_original = False
        force_python_only = False
        memory_factor = 1
        log_frequency = 100

        G0_memmap = SnpMemMap(filename)
        idx = G0_memmap.pos[:,0]!=2
        inonzero = np.array([True]*sum(idx))
        inonzero[0] = False
        logging.info("Generating random {0}x{0}".format(sum(idx)))
        if sum(idx) < 10000:
            np.random.seed(0)
            SVinv3 = np.random.random((sum(idx),sum(idx)))
        else:
            SVinv3 = np.zeros((sum(idx),sum(idx)))
            np.fill_diagonal(SVinv3,val=1)
        logging.info("Done Generating random")
        local_fn_U = r"d:\deldir\local_fn_U.memmap"
        t0 = time.time()
        U_memmap = post_svd(local_fn_U, G0_memmap, idx, SVinv3, inonzero, memory_factor, runner, do_original=do_original,force_python_only=force_python_only,log_frequency=log_frequency)
        print(U_memmap.val)
        logging.info("clocktime {0}".format(format_delta(time.time()-t0)))
        print("done")
        


    elif test_ata:
        if do_fast:
            filename = r"D:\deldir\test\_Storage\very_small_0.8\G0_data.memmap"
        else:
            filename = r"D:\deldir\scratch\escience\carlk\cachebio\genetics\onemil\fc\bigsyn0\U1.memmap"

        work_count = 3
        force_python_only = False

        a = SnpMemMap(filename)
        def memmap_lambda():
            return a
        result = []
        def writer(kernel_data):
            result.append(kernel_data)


        mmultfile_ata(memmap_lambda,writer,a.sid,work_count,name=None,runner=Local(),force_python_only=force_python_only)
        result = result[0]
        print(result.val)
        logging.info("Done")

    else:
        if do_fast:
            u1_filename = r"D:\deldir\scratch\escience\carlk\cachebio\genetics\onemil\fcvery_small_5\U1.memmap"
            test_snps_count = 100
            log_frequency=100
        else:
            u1_filename = r"D:\deldir\scratch\escience\carlk\cachebio\genetics\onemil\fc\bigsyn0\U1.memmap"
            test_snps_count = 1000
            log_frequency=10
        a = SnpMemMap(u1_filename)
        np.random.seed(0)
        logging.info("Generating random {0}x{1}".format(a.iid_count,a.sid_count))
        x = np.random.random((a.iid_count,test_snps_count))
        logging.info("Done Generating random")
        logging.info("Starting C++ run")
        cc1, cc2 = mmultfile_b_less_aatb(u1_filename,x,log_frequency=log_frequency,force_python_only=False)
        logging.info("Starting python run")
        py1, py2 = mmultfile_b_less_aatb(u1_filename,x,log_frequency=log_frequency,force_python_only=True)
        np.testing.assert_array_almost_equal(py1,cc1, decimal=3)
        np.testing.assert_array_almost_equal(py2,cc2, decimal=3)
        logging.info("Done")

